/*
 * Copyright (c) 2000 - 2002 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: bktr.c,v 1.38 2002/01/18 10:55:47 pva Exp $
 *  Driver for cards support for which is compiled into kernel
 */

#include <err.h>
#include <fcntl.h>	/* O_RDONLY */

#include "access.h"

static int stereo = 1; /* Use stereo by default */
static int fd = -1;
static int vol = 0;

#ifdef linux
static double get_freq_fact(int);
static int tuner_ord = 0;
#endif

extern const char *tuner_device_1;
extern const char *tuner_device_2;

/******************************************************************/

int
Grab_Port_bktr(unsigned long port) {
#ifdef BSDBKTR
	int intern = AUDIO_INTERN;
#elif defined linux
	struct video_tuner t;
#endif /* __FreeBSD__ || __OpenBSD__ || __NetBSD__ */

	fd = radio_device_get(tuner_device_1, tuner_device_2, O_RDONLY);
	if (fd < 0)
		return -1;

	/* Check for working driver */
#ifdef BSDBKTR
	if (ioctl(fd, BT848_SAUDIO, &intern) < 0)
#elif defined linux
	/* FIXME: it can be not the first tuner */
	t.tuner = tuner_ord;

	if (ioctl(fd, VIDIOCGTUNER, &t) < 0)
#endif
		return -1;
	return 0;
}

int
Release_Port_bktr(void) {
	return radio_device_release(fd, tuner_device_1);
}

unsigned long
info_port_bktr(void) {
	return 0ul;
}

void
Change_Frequency_bktr(unsigned long frequency) {
#ifdef BSDBKTR
	if (ioctl(fd, RADIO_SETFREQ, &frequency) < 0)
#elif defined linux
	unsigned long freq = (unsigned long)(frequency*get_freq_fact(fd));

	if (ioctl(fd, VIDIOCSFREQ, &freq) < 0)
#endif
		warn("frequency set error");
}

void
mute_bktr(void) {
#ifdef BSDBKTR
	int intern = AUDIO_MUTE;

	if (!vol) return;

	vol = 0;

	if (ioctl(fd, BT848_SAUDIO, &intern) < 0)
#elif defined linux
	struct video_audio va;

	va.audio = 0;
	va.flags = VIDEO_AUDIO_MUTE;
	va.volume = vol = 0;

	if (ioctl(fd, VIDIOCSAUDIO, &va) < 0)
#endif
		warn("mute error");
}

void
volumeup_bktr(void) {
#ifdef BSDBKTR
	int intern = AUDIO_UNMUTE;

	if (vol) return;

	vol = 1;

	if (ioctl(fd, BT848_SAUDIO, &intern) < 0)
		warn("unmute error");
#elif defined linux
	if (vol < 10) {
		struct video_audio va;
		
		va.flags = VIDEO_AUDIO_VOLUME;
		if (stereo)
			va.mode = VIDEO_SOUND_STEREO;
		else
			va.mode = VIDEO_SOUND_MONO;
		va.audio = 0;
		va.volume = (++vol)*(65535/10);

		if (ioctl(fd, VIDIOCSAUDIO, &va) < 0)
			warn("volume up error");
	}
#endif
}

void
volumedn_bktr(void) {
#ifdef BSDBKTR
	int intern = AUDIO_MUTE;

	if (!vol) return;

	vol = 0;

	if (ioctl(fd, BT848_SAUDIO, &intern) < 0)
		warn("mute error");
#elif defined linux
	if (vol > 0) {
		struct video_audio va;

		va.flags = VIDEO_AUDIO_VOLUME;
		if (stereo)
			va.mode = VIDEO_SOUND_STEREO;
		else
			va.mode = VIDEO_SOUND_MONO;
		va.audio = 0;
		va.volume = (--vol)*(65535/10);

		if (ioctl(fd, VIDIOCSAUDIO, &va) < 0)
			warn("volume down error");
	}
#endif
}

int
state_bktr(void) {
#ifdef BSDBKTR
	int ss;

	if (ioctl(fd, TVTUNER_GETSTATUS, &ss) < 0) return 3;

	ss &= 0x07;

	if (ss > 0x06) return 0;
	if (ss > 0x03) return 1;
	if (ss > 0x01) return 2;
#elif defined linux
	struct video_tuner t;

	t.tuner = tuner_ord;

	if (ioctl(fd, VIDIOCGTUNER, &t) < 0) return 3;

	if (t.flags & VIDEO_TUNER_STEREO_ON) return 0; /* STEREO */
	if (t.signal > 49061) return 1; /* almost stereo */
	if (t.signal > 32678) return 2; /* mono */
#endif
	return 3;
}

void
mono_bktr(void) {
#ifdef BSDBKTR
	unsigned int mode = RADIO_MONO;

	if (!stereo) return;

	stereo = 0;

	if (ioctl(fd, RADIO_SETMODE, &mode) < 0)
#elif defined linux
	struct video_audio va;

	va.audio = 0;
	va.mode = VIDEO_SOUND_MONO;
	stereo = 0;

	if (ioctl(fd, VIDIOCSAUDIO, &va) < 0)
#endif
		warn("mono set error");
}

unsigned long
get_freq_bktr(void) {
	unsigned long freq;
#ifdef BSDBKTR
	if (ioctl(fd, RADIO_GETFREQ, &freq) < 0)
		warn("frequency get error");

	return freq;
#else
	float fact = get_freq_fact(fd);

	if (ioctl(fd, VIDIOCGFREQ, &freq) < 0)
		warn("frequency get error");

	if (fact == 160.)
		return (unsigned long)(1 + freq/160);

	return (unsigned long)(freq/fact);
#endif
}

int
get_vol_bktr(void) {
#ifdef BSDBKTR
	int intern;

	if (ioctl(fd, BT848_GAUDIO, &intern) < 0)
		warn("volume get error");

	return intern;
#else
	struct video_audio va;

	va.audio = 0;

	if (ioctl(fd, VIDIOCGAUDIO, &va) < 0)
		warn("volume get error");

	return 1 + va.volume * 10 / 65535;
#endif
}

#ifdef linux
static double
get_freq_fact(int fd) {
	struct video_tuner t;
	t.tuner = tuner_ord;
	if (ioctl(fd, VIDIOCGTUNER, &t) == -1 || (t.flags & VIDEO_TUNER_LOW) == 0)
		return .16;

	return 160.;
}
#endif /* linux */
