/*
 * Copyright (c) 2002 Gunther Mayer
 * Copyright (c 2001, 2002 Vladimir Popov <jumbo@narod.ru>.
 * Copyright (c) 2000, 2001 Wilmer van der Gaast <lintux@lintux.cx>
 * Copyright (c) 1998 Kevin Boone <k.boone@mdx.ac.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* BMC/ FCC ID HMA00-0076 Radio Card driver
 *
 * The PCB has a print "BMC" and "FCC ID HMA00-0076" on it.
 * The card was sold e.g. as some "Typhoon Radio Card" (Note: Typhoon used at least
 *  4 different cards as OEM!)
 *
 * IO is jumper selectable 0x20f or 0x30f
 *
 * Frequency is controlled by
 *   Toshiba's High Speed PLL for DTS TC9216P
 *    3wire protocol pins are connected to Bit 0-2
 * Volume is controlled by
 *   Setting Bits 3-6 (Bit6=MSB)
 *    Note: Volumerange is 0-15; though 12-15 usable, lower than 12 is low volume 
 * Mono/Stereo is controlled by
 *   Bit 7: 1=Stereo 0=Mono
 */

#include "access.h"
#include "tc921x.h"

#define bmc_FREQ_DATA_ON	(1 << 0)
#define bmc_FREQ_CLOCK_ON	(1 << 1)
#define bmc_FREQ_PERIOD_ON	(1 << 2)

#define bmc_VOLUME_MASK		(0x78)  /* Bit 3-6 */

#define BMC_STEREO		(1 << 7) /* 0x80 for Stereo, else: Mono */
#define BMC_MONO		0

#define bmc_FREQ_STEADY	\
		bmc_FREQ_DATA_ON | \
		bmc_FREQ_CLOCK_ON | \
		bmc_FREQ_PERIOD_ON

#define USE_CHANNEL	1

static struct tc921x_t card;
static int vol = 15;
static int stereo = BMC_STEREO;

static void set_vol(int);
static int find_card_bmc(int);

/******************************************************************/

int
Grab_Port_bmc(unsigned long port) {
	card.port = port;

	card.period = bmc_FREQ_PERIOD_ON;
	card.clock = bmc_FREQ_CLOCK_ON;
	card.data = bmc_FREQ_DATA_ON;

	return find_card_bmc(port);
}

int
Release_Port_bmc(void) {
	return radio_release_ioperms(card.port, 1);
}

unsigned long
info_port_bmc(void) {
	return card.port;
}

void
Change_Frequency_bmc(unsigned long frequency) {
	unsigned long data = 0ul;

	data |= tc921x_encode_freq(frequency);
	data |= TC921X_D0_REF_FREQ_10_KHZ;
	data |= TC921X_D0_PULSE_SWALLOW_FM_MODE;
	data |= TC921X_D0_OSC_7POINT2_MHZ;
	data |= TC921X_D0_OUT_CONTROL_ON;
	tc921x_write_addr(&card, 0xD0, data);

	data  = TC921X_D2_IO_PORT_OUTPUT(4);
	tc921x_write_addr(&card, 0xD2, data);

	set_vol(vol);
}

unsigned long
get_freq_bmc(void) {
	return tc921x_decode_freq(tc921x_read_addr(&card, 0xD1));
}

void
mute_bmc(void) {
	vol = 0;
	set_vol(vol);
}

void
volumeup_bmc(void) {
	if (vol < 15)
		vol++;

	set_vol(vol);
}

void
volumedn_bmc(void) {
	if (vol > 0)
		vol--;

	set_vol(vol);
}

int
state_bmc(void) {
	return 3;
}

static void
set_vol(int v) {
	int i;

	vol = v;	
	i = (v & 0x0f) << 3; /* set volume bits */
	i |= stereo | 0x07; /* leave 3wire tc9216 bits HI, else undef state */
	OUTB(card.port, i);
}

static int
find_card_bmc(int radioport) {
        unsigned long cur_freq = 0ul;
        unsigned long test_freq = 0ul;
        int err = -1;

        if (radio_get_ioperms(radioport, 1) < 0)
                return -1;

        cur_freq = get_freq_bmc();

        Change_Frequency_bmc(TEST_FREQ);
        test_freq = get_freq_bmc();
        if (test_freq == TEST_FREQ) {
                Change_Frequency_bmc(cur_freq);
                err = 0;
        }

        return err;
}

void set_mono_bmc()
{
	stereo = BMC_MONO;
	set_vol(vol);
}

