/*
 * Copyright (c) 2001, 2002 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: sf16fmr2.c,v 1.38 2002/01/18 10:55:51 pva Exp $
 * lowlevel driver io functions for fmio utility -- sf16fmr2 driver
 */

#include "access.h"
#include "pt2254a.h"
#include "tea5757.h"

#define SF16FMR2_VOLU_STROBE_ON  (0 << 2)
#define SF16FMR2_VOLU_STROBE_OFF (1 << 2)
#define SF16FMR2_VOLU_CLOCK_ON   (1 << 5)
#define SF16FMR2_VOLU_CLOCK_OFF  (0 << 5)
#define SF16FMR2_VOLU_DATA_ON    (1 << 6)
#define SF16FMR2_VOLU_DATA_OFF   (0 << 6)

static struct tea5757_t card;
static unsigned long radioport = 0;
static int type = 0;
static int volume = 0;

static void write_shift_register(unsigned long);
static unsigned long read_shift_register(void);
static int find_card_sf16fmr2(void);
static void set_volume(int);
static void send_vol_bit(int);

/******************************************************************/

int
Grab_Port_sf16fmr2(unsigned long port) {
	card.frequency = 0;
	card.search = TEA5757_SEARCH_END;
	card.sensitivity = TEA5757_S030;
	card.stereo = TEA5757_STEREO;
	card.write = write_shift_register;
	card.read = read_shift_register;
	radioport = 0x384;

	return find_card_sf16fmr2();
}

int
Release_Port_sf16fmr2(void) {
	return radio_release_ioperms(radioport, 1);
}

unsigned long
info_port_sf16fmr2(void) {
	return radioport;
}

void
set_frequency_sf16fmr2(unsigned long frequency) {
	card.frequency = frequency;
	card.search = TEA5757_SEARCH_END;
	tea5757_write_shift_register(&card);
	return;
}

unsigned long
search_up_sf16fmr2(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_UP;
	return tea5757_search(&card);
}

unsigned long
search_down_sf16fmr2(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_DOWN;
	return tea5757_search(&card);
}

void
mute_sf16fmr2(void) {
	OUTB(radioport, 0x00);
}

void
unmute_sf16fmr2(void) {
	OUTB(radioport, 0x04);
}

void
volumeup_sf16fmr2(void) {
	unmute_sf16fmr2();
	if (type == 0) {
		volume = 1;
	} else {
		if (volume < 15)
			volume++;

		set_volume(volume);
	}
}

void
volumedown_sf16fmr2(void) {
	if (type == 0) {
		mute_sf16fmr2();
		volume = 0;
	} else {
		if (volume == 0) {
			mute_sf16fmr2();
			return;
		}

		if (volume > 0)
			volume--;

		set_volume(volume);
	}
}

int
state_sf16fmr2(void) {
	unsigned int res = tea5757_read_shift_register(&card);
	return ((res >> 25) ^ 0x03) & 0x03;
}

void
mono_sf16fmr2(void) {
	card.stereo = TEA5757_MONO;
}

static unsigned long
read_shift_register(void) {
	int rb, res;
	int state;

	OUTB(radioport, 0x05);

	OUTB(radioport, 0x07);
	rb = inb(radioport);
	state = rb & 0x80 ? 0x04 : 0; /* Amplifier present/not present */
	type = rb & 0x80 ? 1 : 0;
	state |= rb & 0x08 ? 0 : 0x02; /* Tuned/Not tuned */

	OUTB(radioport, 0x05);
	rb = inb(radioport);
	state |= rb & 0x08 ? 0 : 0x01; /* Mono/Stereo */
	res = rb & 0x01;

	/* Read the rest of the register */
	rb = 23;
	while (rb--) {
		res <<= 1;
		OUTB(radioport, 0x07);
		OUTB(radioport, 0x05);
		res |= inb(radioport) & 0x01;
	}

	return res | (state << 25);
}

static void
write_shift_register(unsigned long data) {
	int c = 25;

	OUTB(radioport, 0x00);

	while (c--)
		if (data & (1 << c)) {
			OUTB(radioport, 0x01);
			OUTB(radioport, 0x03);
		} else {
			OUTB(radioport, 0x00);
			OUTB(radioport, 0x02);
		}

	OUTB(radioport, 0x00);
	OUTB(radioport, 0x04);
}

unsigned long
get_frequency_sf16fmr2(void) {
	return tea5757_decode_frequency(tea5757_read_shift_register(&card));
}

static int
find_card_sf16fmr2(void) {
	unsigned long cur_freq = 0ul;
	unsigned long test_freq = 0ul;
	int err = -1;

	if (radio_get_ioperms(radioport, 1) < 0)
		return -1;
	cur_freq = get_frequency_sf16fmr2();

	set_frequency_sf16fmr2(TEST_FREQ);
	test_freq = get_frequency_sf16fmr2();
	if (test_freq == TEST_FREQ) {
		set_frequency_sf16fmr2(cur_freq);
		err = 0;
	}

	return err;
}

void
set_volume(int v) {
	unsigned long reg, vol;
	int i;

	vol = pt2254a_encode_volume(v, 15);
	reg = pt2254a_compose_register(vol, vol, USE_CHANNEL, USE_CHANNEL);

	OUTB(radioport, SF16FMR2_VOLU_STROBE_OFF);

	for (i = 0; i < PT2254A_REGISTER_LENGTH; i++)
		send_vol_bit(reg & (1 << i));

	/* Latch the data */
	OUTB(radioport, SF16FMR2_VOLU_STROBE_ON);
	OUTB(radioport, SF16FMR2_VOLU_STROBE_OFF);
	OUTB(radioport, 0x10 | SF16FMR2_VOLU_STROBE_OFF);
}

static void
send_vol_bit(int i) {
	unsigned int data;

	data  = SF16FMR2_VOLU_STROBE_OFF;
	data |= i ? SF16FMR2_VOLU_DATA_ON : SF16FMR2_VOLU_DATA_OFF;

	OUTB(radioport, data | SF16FMR2_VOLU_CLOCK_OFF);
	OUTB(radioport, data | SF16FMR2_VOLU_CLOCK_ON);
}
