/*
 * Copyright (c) 2002 Vladimir Popov <jumbo@narod.ru>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * fmrinit - this program initializes SF16-FMR cards. Needed only once at boot.
 */


#include <sys/param.h>

#include <err.h>

#ifdef __FreeBSD__
#include <machine/cpufunc.h>
#include <fcntl.h>
#include <unistd.h>
#elif defined __OpenBSD__ || defined __NetBSD__
#include <machine/sysarch.h>
#elif defined linux
#include <sys/io.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../tc921x.h"

#define SF16FMR_FREQ_DATA_ON	(1 << 0)
#define SF16FMR_FREQ_CLOCK_ON	(1 << 1)
#define SF16FMR_FREQ_PERIOD_ON	(1 << 2)

#define SF16FMR_0x284		0x284
#define SF16FMR_0x384		0x384

#define DEF_PORT		SF16FMR_0x384

#ifdef __FreeBSD__
const char *devio = "/dev/io";
static int fd = -1;

int
fbsd_get_ioperms(void) {
	if ((fd = open(devio, O_RDONLY)) < 0) {
		warn("%s open error", devio);
		return -1;
	}

	return 0;
}

int
fbsd_release_ioperms(void) {
	if (close(fd) < 0) {
		warn("%s close error", devio);
		return -1;
	}

	return 0;
}
#elif defined linux
int
os_ioperms(unsigned long port, int no, int v) {
	if (ioperm(port, no, v) < 0) {
		warn("ioperm() error");
		return -1;
	}

	return 0;
}
#elif defined __OpenBSD__ || defined __NetBSD__
int
os_ioperms(unsigned long port, int no, int v) {
	unsigned long iomap[32];
	int offset;
	unsigned long mask;
	struct i386_set_ioperm_args ioperms;

	memset(iomap, 0xFFFF, sizeof(iomap));
	if (v) {
		offset = port / 32;
		mask = no << port % 32;
		iomap[offset] ^= mask;
        }
	ioperms.iomap = iomap;
	if (sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0) {
		warn("I386_SET_IOPERM error");
		return -1;
	}

	return 0;
}
#else
int
os_ioperms(unsigned long port, int no, int v) {
	return -1;
}
#endif /* __FreeBSD__ */

int
radio_get_ioperms(unsigned long port, int no) {
#ifdef __FreeBSD__
	return fbsd_get_ioperms();
#else
	return os_ioperms(port, no, 1);
#endif
}

int
radio_release_ioperms(unsigned long port, int no) {
#ifdef __FreeBSD__
	return fbsd_release_ioperms();
#else
	return os_ioperms(port, no, 0);
#endif
}

int
main(int argc, char **argv) {
	struct tc921x_t card;
	unsigned long data = 0ul;

	card.port = DEF_PORT;

	if (argc > 1)
		sscanf(argv[1], "%i", (int *)&card.port);

	if (card.port != SF16FMR_0x284 && card.port != SF16FMR_0x384)
		card.port = DEF_PORT;

	if (radio_get_ioperms(card.port, 1) < 0)
		return 1;

	card.period = SF16FMR_FREQ_PERIOD_ON;
	card.clock = SF16FMR_FREQ_CLOCK_ON;
	card.data = SF16FMR_FREQ_DATA_ON;

	/* Do a dummy data transfer to initialize the chip */
	data |= TC921X_D0_REF_FREQ_10_KHZ;
	data |= TC921X_D0_PULSE_SWALLOW_FM_MODE;
	data |= TC921X_D0_OSC_7POINT2_MHZ;
	data |= TC921X_D0_OUT_CONTROL_ON;
	tc921x_write_addr(&card, 0xD0, data);

	data  = TC921X_D2_IO_PORT_OUTPUT(4);
	tc921x_write_addr(&card, 0xD2, data);

	if (radio_release_ioperms(card.port, 1) < 0)
		return 1;

	return 0;
}
